@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2016 Andrew M. Bishop
+ This file Copyright 2008-2017, 2019, 2024 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -129,6 +129,9 @@ Results *CalculateRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *re
        prev_segment=FindClosestSegment(nodes,segments,ways,lat,lon,1,profile,&distmin,&node1,&node2,&dist1,&dist2);
       }
 
+    if(IsFakeSegment(prev_segment))
+       prev_segment=IndexRealSegment(prev_segment);
+
     fake_segment=CreateFakeNullSegment(segments,start_node,prev_segment,finish_waypoint);
 
     result1=InsertResult(complete,start_node,prev_segment);
@@ -318,7 +321,7 @@ static Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
    {
     Node *startp=LookupNode(nodes,start_node,1);
 
-    if(!(startp->allow&profile->allow))
+    if(!(startp->allow&profile->transports))
        force_uturn=1;
    }
 
@@ -374,7 +377,7 @@ static Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
        /* must obey one-way restrictions (unless profile allows) */
        if(profile->oneway && IsOnewayTo(segment2p,node1))
          {
-          if(profile->allow!=Transports_Bicycle)
+          if(profile->transports!=Transports_Bicycle)
              goto endloop;
 
           way2p=LookupWay(ways,segment2p->way,1);
@@ -406,7 +409,7 @@ static Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
              goto endloop;
 
        /* must obey turn relations */
-       if(turnrelation!=NO_RELATION && !IsTurnAllowed(relations,turnrelation,node1,seg1r,seg2r,profile->allow))
+       if(turnrelation!=NO_RELATION && !IsTurnAllowed(relations,turnrelation,node1,seg1r,seg2r,profile->transports))
           goto endloop;
 
        if(!IsFakeNode(node2))
@@ -419,7 +422,7 @@ static Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
        way2p=LookupWay(ways,segment2p->way,1);
 
        /* mode of transport must be allowed on the highway */
-       if(!(way2p->allow&profile->allow))
+       if(!(way2p->allow&profile->transports))
           goto endloop;
 
        /* must obey weight restriction (if exists) */
@@ -439,7 +442,7 @@ static Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
           goto endloop;
 
        for(i=1;i<Property_Count;i++)
-          if(ways->file.props & PROPERTIES(i))
+          if(ways->file.properties & PROPERTIES(i))
             {
              if(way2p->props & PROPERTIES(i))
                 segment_pref*=profile->props_yes[i];
@@ -452,7 +455,7 @@ static Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
           goto endloop;
 
        /* mode of transport must be allowed through node2 unless it is the final node */
-       if(node2p && node2!=finish_node && !(node2p->allow&profile->allow))
+       if(node2p && node2!=finish_node && !(node2p->allow&profile->transports))
           goto endloop;
 
        /* calculate the score for the segment and cumulative */
@@ -628,7 +631,7 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
       {
        Node *startp=LookupNode(nodes,begin->start_node,1);
 
-       if(!(startp->allow&profile->allow))
+       if(!(startp->allow&profile->transports))
           force_uturn=1;
       }
    }
@@ -722,6 +725,8 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
 
  while(1)
    {
+    int queue1_empty=0,queue2_empty=0;
+
     /* Forward queue */
 
     if((result1=PopFromQueue(fwd_queue)))
@@ -765,7 +770,7 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
           /* must obey one-way restrictions (unless profile allows) */
           if(profile->oneway && IsOnewayTo(segment2p,node1))
             {
-             if(profile->allow!=Transports_Bicycle)
+             if(profile->transports!=Transports_Bicycle)
                 goto endloop_fwd;
 
              way2p=LookupWay(ways,segment2p->way,1);
@@ -788,13 +793,13 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
                 goto endloop_fwd;
 
           /* must obey turn relations */
-          if(turnrelation!=NO_RELATION && !IsTurnAllowed(relations,turnrelation,node1,seg1,seg2,profile->allow))
+          if(turnrelation!=NO_RELATION && !IsTurnAllowed(relations,turnrelation,node1,seg1,seg2,profile->transports))
              goto endloop_fwd;
 
           way2p=LookupWay(ways,segment2p->way,1);
 
           /* mode of transport must be allowed on the highway */
-          if(!(way2p->allow&profile->allow))
+          if(!(way2p->allow&profile->transports))
              goto endloop_fwd;
 
           /* must obey weight restriction (if exists) */
@@ -814,7 +819,7 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
              goto endloop_fwd;
 
           for(i=1;i<Property_Count;i++)
-             if(ways->file.props & PROPERTIES(i))
+             if(ways->file.properties & PROPERTIES(i))
                {
                 if(way2p->props & PROPERTIES(i))
                    segment_pref*=profile->props_yes[i];
@@ -831,7 +836,7 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
           node2p=LookupNode(nodes,node2,2); /* node2 cannot be a fake node (must be a super-node) */
 
           /* mode of transport must be allowed through node2 unless it is the final node */
-          if(node2!=end->finish_node && !(node2p->allow&profile->allow))
+          if(node2!=end->finish_node && !(node2p->allow&profile->transports))
              goto endloop_fwd;
 
           /* calculate the score for the segment and cumulative */
@@ -851,11 +856,11 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
 
           if(result2 && result2->next)
             {
-             if((cumulative_score+result2->score)<total_score && !result2->prev)
+             if((result2->score+cumulative_score)<total_score)
                {
-                result2->prev=result1;
-                total_score=cumulative_score+result2->score;
-                finish_result=start_result=result2;
+                total_score=result2->score+cumulative_score;
+                finish_result=result2;
+                start_result =result1;
                }
 
              goto endloop_fwd;
@@ -904,11 +909,11 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
 #endif
       }
     else
-       break;
+       queue1_empty=1;
 
     /* Reverse queue */
 
-    if(!finish_result && (result1=PopFromQueue(rev_queue)))
+    if((result1=PopFromQueue(rev_queue)))
       {
        Node *node1p;
        Segment *segment1p,*segment2p;
@@ -931,7 +936,7 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
        node1p=LookupNode(nodes,node1,1);
 
        /* mode of transport must be allowed through node1 */
-       if(!(node1p->allow&profile->allow))
+       if(!(node1p->allow&profile->transports))
           continue;
 
        way1p=LookupWay(ways,segment1p->way,1);
@@ -939,7 +944,7 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
        segment1_pref=profile->highway[HIGHWAY(way1p->type)];
 
        for(i=1;i<Property_Count;i++)
-          if(ways->file.props & PROPERTIES(i))
+          if(ways->file.properties & PROPERTIES(i))
             {
              if(way1p->props & PROPERTIES(i))
                 segment1_pref*=profile->props_yes[i];
@@ -984,17 +989,14 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
             {
              index_t turnrelation2=FindFirstTurnRelation2(relations,node1,seg2);
 
-             if(turnrelation2!=NO_RELATION && !IsTurnAllowed(relations,turnrelation2,node1,seg2,seg1,profile->allow))
+             if(turnrelation2!=NO_RELATION && !IsTurnAllowed(relations,turnrelation2,node1,seg2,seg1,profile->transports))
                 goto endloop_rev;
             }
 
-          if(result2 && result2->prev)
-             goto endloop_rev;
-
           /* must obey one-way restrictions (unless profile allows) */
           if(profile->oneway && IsOnewayFrom(segment2p,node1)) /* working backwards => disallow oneway *from* node1 */
             {
-             if(profile->allow!=Transports_Bicycle)
+             if(profile->transports!=Transports_Bicycle)
                 goto endloop_rev;
 
              way2p=LookupWay(ways,segment2p->way,1);
@@ -1006,7 +1008,7 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
           way2p=LookupWay(ways,segment2p->way,1);
 
           /* mode of transport must be allowed on the highway */
-          if(!(way2p->allow&profile->allow))
+          if(!(way2p->allow&profile->transports))
              goto endloop_rev;
 
           /* must obey weight restriction (if exists) */
@@ -1026,7 +1028,7 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
              goto endloop_rev;
 
           for(i=1;i<Property_Count;i++)
-             if(ways->file.props & PROPERTIES(i))
+             if(ways->file.properties & PROPERTIES(i))
                {
                 if(way2p->props & PROPERTIES(i))
                    segment_pref*=profile->props_yes[i];
@@ -1040,14 +1042,24 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
 
           node2=OtherNode(segment2p,node1);
 
-          node2p=LookupNode(nodes,node2,2); /* node2 cannot be a fake node (must be a super-node) */
-
           cumulative_score=result1->score+segment1_score;
 
           /* score must be better than current best score */
           if(cumulative_score>=total_score)
              goto endloop_rev;
 
+          if(result2 && result2->prev)
+            {
+             if((result2->score+cumulative_score)<total_score)
+               {
+                total_score=result2->score+cumulative_score;
+                finish_result=result1;
+                start_result =result2;
+               }
+
+             goto endloop_rev;
+            }
+
           if(!result2) /* New end node/segment pair */
             {
              result2=InsertResult(results,node1,seg2); /* adding in reverse => node1,seg2 */
@@ -1064,7 +1076,9 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
 
           /* Insert a new node into the queue */
 
-          GetLatLong(nodes,node2,node2p,&lat,&lon); /* node2 cannot be a fake node (must be a super-node) */
+          node2p=LookupNode(nodes,node2,2); /* node2 cannot be a fake node (must be a super-node) */
+
+          GetLatLong(nodes,node2,node2p,&lat,&lon);
 
           direct=Distance(lat,lon,start_lat,start_lon);
 
@@ -1090,7 +1104,10 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
             }
 #endif
       }
-    else if(!finish_result)
+    else
+       queue2_empty=1;
+
+    if(queue1_empty || queue2_empty)
        break;
    }
 
@@ -1116,21 +1133,22 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
 
  /* Turn the route round and fill in the start and finish information */
 
- result1=FirstResult(results);
+ if(start_result!=finish_result)
+   {
+    start_result->next=finish_result;
+    finish_result->prev=start_result;
 
- while(result1)
-    result1=NextResult(results,result1);
+    while(start_result->prev && start_result->prev!=NO_RESULT)
+       start_result=start_result->prev;
 
- while(start_result->prev && start_result->prev!=NO_RESULT)
-    start_result=start_result->prev;
+    FixForwardRoute(results,finish_result);
 
- FixForwardRoute(results,finish_result);
+    if(!start_result->prev && start_result->next)
+       start_result=start_result->next;
 
- if(!start_result->prev && start_result->next)
-    start_result=start_result->next;
-
- while(finish_result->next && finish_result->next!=NO_RESULT)
-    finish_result=finish_result->next;
+    while(finish_result->next && finish_result->next!=NO_RESULT)
+       finish_result=finish_result->next;
+   }
 
  results->start_node=start_result->node;
  results->prev_segment=start_result->segment;
@@ -1139,7 +1157,7 @@ static Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relat
  results->last_segment=finish_result->segment;
 
 #if DEBUG
- printf("    -------- middle route (via super-nodes/segments)\n");
+ printf("    -------- middle route (via super-nodes/segments) score=%.3f\n",total_score);
 
  print_debug_route(nodes,segments,results,NULL,4,+1);
 #endif
@@ -1316,7 +1334,7 @@ static Results *FindSuperRoute(Nodes *nodes,Segments *segments,Ways *ways,Relati
          {
           Way *way2p;
 
-          if(profile->allow!=Transports_Bicycle)
+          if(profile->transports!=Transports_Bicycle)
              goto endloop;
 
           way2p=LookupWay(ways,segment2p->way,2);
@@ -1446,7 +1464,7 @@ static Results *FindStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relat
    {
     Node *startp=LookupNode(nodes,start_node,1);
 
-    if(!(startp->allow&profile->allow))
+    if(!(startp->allow&profile->transports))
        force_uturn=1;
    }
 
@@ -1502,7 +1520,7 @@ static Results *FindStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relat
        /* must obey one-way restrictions (unless profile allows) */
        if(profile->oneway && IsOnewayTo(segment2p,node1))
          {
-          if(profile->allow!=Transports_Bicycle)
+          if(profile->transports!=Transports_Bicycle)
              goto endloop;
 
           way2p=LookupWay(ways,segment2p->way,1);
@@ -1534,13 +1552,13 @@ static Results *FindStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relat
              goto endloop;
 
        /* must obey turn relations */
-       if(turnrelation!=NO_RELATION && !IsTurnAllowed(relations,turnrelation,node1,seg1r,seg2r,profile->allow))
+       if(turnrelation!=NO_RELATION && !IsTurnAllowed(relations,turnrelation,node1,seg1r,seg2r,profile->transports))
           goto endloop;
 
        way2p=LookupWay(ways,segment2p->way,1);
 
        /* mode of transport must be allowed on the highway */
-       if(!(way2p->allow&profile->allow))
+       if(!(way2p->allow&profile->transports))
           goto endloop;
 
        /* must obey weight restriction (if exists) */
@@ -1560,7 +1578,7 @@ static Results *FindStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relat
           goto endloop;
 
        for(i=1;i<Property_Count;i++)
-          if(ways->file.props & PROPERTIES(i))
+          if(ways->file.properties & PROPERTIES(i))
             {
              if(way2p->props & PROPERTIES(i))
                 segment_pref*=profile->props_yes[i];
@@ -1576,7 +1594,7 @@ static Results *FindStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relat
           node2p=LookupNode(nodes,node2,2);
 
        /* mode of transport must be allowed through node2 unless it is the final node */
-       if(node2p && node2!=finish_node && !(node2p->allow&profile->allow))
+       if(node2p && node2!=finish_node && !(node2p->allow&profile->transports))
           goto endloop;
 
        /* calculate the score for the segment and cumulative */
@@ -1761,7 +1779,7 @@ static Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Rela
  while((result1=PopFromQueue(queue)))
    {
     Node *node1p=NULL;
-    Segment *segment1p,*segment2p;
+    Segment *segment1p=NULL,*segment2p;
     Way *way1p;
     index_t real_node1,node1,seg1,seg1r;
     index_t turnrelation=NO_RELATION;
@@ -1794,7 +1812,7 @@ static Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Rela
 
     /* mode of transport must be allowed through node1 */
     if(seg1!=NO_SEGMENT)
-       if(node1p && !(node1p->allow&profile->allow))
+       if(node1p && !(node1p->allow&profile->transports))
           continue;
 
     if(seg1!=NO_SEGMENT)
@@ -1804,7 +1822,7 @@ static Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Rela
        segment1_pref=profile->highway[HIGHWAY(way1p->type)];
 
        for(i=1;i<Property_Count;i++)
-          if(ways->file.props & PROPERTIES(i))
+          if(ways->file.properties & PROPERTIES(i))
             {
              if(way1p->props & PROPERTIES(i))
                 segment1_pref*=profile->props_yes[i];
@@ -1848,7 +1866,7 @@ static Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Rela
        /* must obey one-way restrictions (unless profile allows) */
        if(profile->oneway && IsOnewayFrom(segment2p,node1)) /* working backwards => disallow oneway *from* node1 */
          {
-          if(profile->allow!=Transports_Bicycle)
+          if(profile->transports!=Transports_Bicycle)
              goto endloop;
 
           way2p=LookupWay(ways,segment2p->way,1);
@@ -1894,14 +1912,14 @@ static Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Rela
              turnrelation=FindFirstTurnRelation2(relations,node1,seg2r);
 
           /* must obey turn relations */
-          if(turnrelation!=NO_RELATION && !IsTurnAllowed(relations,turnrelation,node1,seg2r,seg1r,profile->allow))
+          if(turnrelation!=NO_RELATION && !IsTurnAllowed(relations,turnrelation,node1,seg2r,seg1r,profile->transports))
              goto endloop;
          }
 
        way2p=LookupWay(ways,segment2p->way,1);
 
        /* mode of transport must be allowed on the highway */
-       if(!(way2p->allow&profile->allow))
+       if(!(way2p->allow&profile->transports))
           goto endloop;
 
        /* must obey weight restriction (if exists) */
@@ -1921,7 +1939,7 @@ static Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Rela
           goto endloop;
 
        for(i=1;i<Property_Count;i++)
-          if(ways->file.props & PROPERTIES(i))
+          if(ways->file.properties & PROPERTIES(i))
             {
              if(way2p->props & PROPERTIES(i))
                 segment_pref*=profile->props_yes[i];
@@ -1937,7 +1955,7 @@ static Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Rela
           node2p=LookupNode(nodes,node2,2);
 
        /* mode of transport must be allowed through node2 */
-       if(node2p && !(node2p->allow&profile->allow))
+       if(node2p && !(node2p->allow&profile->transports))
           goto endloop;
 
        cumulative_score=result1->score+segment1_score;
