@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2013 Andrew M. Bishop
+ This file Copyright 2008-2015, 2018, 2023 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -45,26 +45,33 @@
 /*+ Checks if a value in the XML is one of the allowed values for false. +*/
 #define ISFALSE(xx) (!strcmp(xx,"false") || !strcmp(xx,"no") || !strcmp(xx,"0"))
 
-/* Local variables */
+/* Local parsing variables (re-initialised for each file) */
 
 static NodesX     *nodes;
 static WaysX      *ways;
 static RelationsX *relations;
 
-static node_t *way_nodes=NULL;
-static int     way_nnodes=0;
+static node_t     *way_nodes;
+static int         way_nnodes;
 
-static node_t     *relation_nodes=NULL;
-static int         relation_nnodes=0;
-static way_t      *relation_ways=NULL;
-static int         relation_nways=0;
-static relation_t *relation_relations=NULL;
-static int         relation_nrelations=0;
-static way_t       relation_from=NO_WAY_ID;
-static way_t       relation_to=NO_WAY_ID;
-static node_t      relation_via=NO_NODE_ID;
+static node_t     *relation_nodes;
+static int         relation_nnodes;
+static way_t      *relation_ways;
+static int         relation_nways;
+static relation_t *relation_relations;
+static int         relation_nrelations;
+static way_t       relation_from;
+static int         relation_from_count;
+static way_t       relation_to;
+static int         relation_to_count;
+static node_t      relation_via;
+static int         relation_via_count;
 
-/* Local functions */
+/* Local parsing functions */
+
+static void ProcessNodeTags(TagList *tags,int64_t node_id,double latitude,double longitude,int mode);
+static void ProcessWayTags(TagList *tags,int64_t way_id, int mode);
+static void ProcessRelationTags(TagList *tags,int64_t relation_id,int mode);
 
 static double parse_speed(way_t id,const char *k,const char *v);
 static double parse_weight(way_t id,const char *k,const char *v);
@@ -161,8 +168,12 @@ void AddRelationRefs(int64_t node_id,int64_t way_id,int64_t relation_id,const ch
     relation_nrelations=0;
 
     relation_from=NO_WAY_ID;
-    relation_via=NO_NODE_ID;
+    relation_from_count=0;
+    relation_from_count=0;
     relation_to=NO_WAY_ID;
+    relation_to_count=0;
+    relation_via=NO_NODE_ID;
+    relation_via_count=0;
    }
  else if(node_id!=0)
    {
@@ -179,7 +190,11 @@ void AddRelationRefs(int64_t node_id,int64_t way_id,int64_t relation_id,const ch
     if(role)
       {
        if(!strcmp(role,"via"))
-          relation_via=id;
+         {
+          relation_via_count++;
+          if(relation_via==NO_NODE_ID)
+             relation_via=id;
+         }
       }
    }
  else if(way_id!=0)
@@ -197,9 +212,17 @@ void AddRelationRefs(int64_t node_id,int64_t way_id,int64_t relation_id,const ch
     if(role)
       {
        if(!strcmp(role,"from"))
-          relation_from=id;
+         {
+          relation_from_count++;
+          if(relation_from==NO_WAY_ID)
+             relation_from=id;
+         }
        else if(!strcmp(role,"to"))
-          relation_to=id;
+         {
+          relation_to_count++;
+          if(relation_to==NO_WAY_ID)
+             relation_to=id;
+         }
       }
    }
  else /* if(relation_id!=0) */
@@ -217,6 +240,73 @@ void AddRelationRefs(int64_t node_id,int64_t way_id,int64_t relation_id,const ch
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Add a node after processing the tags associated with it.
+
+  int64_t node_id The id of the node.
+
+  double latitude The latitude of the node.
+
+  double longitude The longitude of the node.
+
+  int mode The mode of operation to take (create, modify, delete).
+
+  TagList *raw_tags The accumulated set of tags for the node.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void AddNode(int64_t node_id,double latitude,double longitude,int mode,TagList *raw_tags)
+{
+ TagList *processed_tags=ApplyNodeTaggingRules(raw_tags,node_id);
+
+ ProcessNodeTags(processed_tags,node_id,latitude,longitude,mode);
+
+ DeleteTagList(raw_tags);
+ DeleteTagList(processed_tags);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Add a way after processing the tags associated with it.
+
+  int64_t way_id The id of the way.
+
+  int mode The mode of operation to take (create, modify, delete).
+
+  TagList *raw_tags The accumulated set of tags for the node.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void AddWay(int64_t way_id,int mode,TagList *raw_tags)
+{
+ TagList *processed_tags=ApplyWayTaggingRules(raw_tags,way_id);
+
+ ProcessWayTags(processed_tags,way_id,mode);
+
+ DeleteTagList(raw_tags);
+ DeleteTagList(processed_tags);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Add a relation after processing the tags associated with it.
+
+  int64_t relation_id The id of the relation.
+
+  int mode The mode of operation to take (create, modify, delete).
+
+  TagList *raw_tags The accumulated set of tags for the node.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void AddRelation(int64_t relation_id,int mode,TagList *raw_tags)
+{
+ TagList *processed_tags=ApplyRelationTaggingRules(raw_tags,relation_id);
+
+ ProcessRelationTags(processed_tags,relation_id,mode);
+
+ DeleteTagList(raw_tags);
+ DeleteTagList(processed_tags);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Process the tags associated with a node.
 
@@ -231,7 +321,7 @@ void AddRelationRefs(int64_t node_id,int64_t way_id,int64_t relation_id,const ch
   int mode The mode of operation to take (create, modify, delete).
   ++++++++++++++++++++++++++++++++++++++*/
 
-void ProcessNodeTags(TagList *tags,int64_t node_id,double latitude,double longitude,int mode)
+static void ProcessNodeTags(TagList *tags,int64_t node_id,double latitude,double longitude,int mode)
 {
  transports_t allow=Transports_ALL;
  nodeflags_t flags=0;
@@ -409,11 +499,11 @@ void ProcessNodeTags(TagList *tags,int64_t node_id,double latitude,double longit
   int mode The mode of operation to take (create, modify, delete).
   ++++++++++++++++++++++++++++++++++++++*/
 
-void ProcessWayTags(TagList *tags,int64_t way_id,int mode)
+static void ProcessWayTags(TagList *tags,int64_t way_id,int mode)
 {
  Way way={0};
  int oneway=0,area=0;
- int roundabout=0,lanes=0;
+ int roundabout=0,lanes=0,cyclebothways=0;
  char *name=NULL,*ref=NULL,*refname=NULL;
  way_t id;
  int i;
@@ -525,6 +615,17 @@ void ProcessWayTags(TagList *tags,int64_t way_id,int mode)
 
        break;
 
+      case 'c':
+       if(!strcmp(k,"cyclebothways"))
+         {
+          if(ISTRUE(v))
+             cyclebothways=1;
+          else if(!ISFALSE(v))
+             logerror("Way %"Pway_t" has an unrecognised tag 'cyclebothways' = '%s' (after tagging rules); using 'no'.\n",logerror_way(id),v);
+          recognised=1; break;
+         }
+       break;
+
       case 'f':
        if(!strcmp(k,"foot"))
          {
@@ -766,13 +867,28 @@ void ProcessWayTags(TagList *tags,int64_t way_id,int mode)
 
  if(area && oneway)
    {
-    logerror("Way %"Pway_t" is an area and oneway; ignoring area tagging.\n",logerror_way(id));
+    logerror("Way %"Pway_t" is an area and oneway; ignoring area tag.\n",logerror_way(id));
     area=0;
    }
 
+ if(cyclebothways && !oneway)
+   {
+    logerror("Way %"Pway_t" is cyclebothways but not oneway; ignoring cyclebothways tag.\n",logerror_way(id));
+    cyclebothways=0;
+   }
+
+ if(roundabout && !oneway)
+   {
+    logerror("Way %"Pway_t" is roundabout but not oneway; adding oneway tag.\n",logerror_way(id));
+    oneway=1;
+   }
+
  if(!way.allow)
     return;
 
+ if(cyclebothways)
+    way.type|=Highway_CycleBothWays;
+
  if(oneway)
    {
     way.type|=Highway_OneWay;
@@ -837,7 +953,7 @@ void ProcessWayTags(TagList *tags,int64_t way_id,int mode)
   int mode The mode of operation to take (create, modify, delete).
   ++++++++++++++++++++++++++++++++++++++*/
 
-void ProcessRelationTags(TagList *tags,int64_t relation_id,int mode)
+static void ProcessRelationTags(TagList *tags,int64_t relation_id,int mode)
 {
  transports_t routes=Transports_None;
  transports_t except=Transports_None;
@@ -983,21 +1099,43 @@ void ProcessRelationTags(TagList *tags,int64_t relation_id,int mode)
        /* Extra logerror information since relation isn't stored */
        if(relation_to!=NO_WAY_ID) logerror_way(relation_to);
        if(relation_via!=NO_NODE_ID) logerror_node(relation_via);
-       logerror("Relation %"Prelation_t" is a turn restriction but has no 'from' way.\n",logerror_relation(id));
+       logerror("Turn Relation %"Prelation_t" has no 'from' way.\n",logerror_relation(id));
       }
     if(relation_to==NO_WAY_ID)
       {
        /* Extra logerror information since relation isn't stored */
        if(relation_via!=NO_NODE_ID) logerror_node(relation_via);
        if(relation_from!=NO_WAY_ID) logerror_way(relation_from);
-       logerror("Relation %"Prelation_t" is a turn restriction but has no 'to' way.\n",logerror_relation(id));
+       logerror("Turn Relation %"Prelation_t" has no 'to' way.\n",logerror_relation(id));
       }
     if(relation_via==NO_NODE_ID)
       {
        /* Extra logerror information since relation isn't stored */
        if(relation_to!=NO_WAY_ID) logerror_way(relation_to);
        if(relation_from!=NO_WAY_ID) logerror_way(relation_from);
-       logerror("Relation %"Prelation_t" is a turn restriction but has no 'via' node.\n",logerror_relation(id));
+       logerror("Turn Relation %"Prelation_t" has no 'via' node.\n",logerror_relation(id));
+      }
+
+    if(relation_from_count>1)
+      {
+       if(relation_to!=NO_WAY_ID) logerror_way(relation_to);
+       if(relation_via!=NO_NODE_ID) logerror_node(relation_via);
+       if(relation_from!=NO_WAY_ID) logerror_way(relation_from);
+       logerror("Turn Relation %"Prelation_t" has more than one 'from' Way (used first one).\n",logerror_relation(relation_id));
+      }
+    if(relation_to_count>1)
+      {
+       if(relation_to!=NO_WAY_ID) logerror_way(relation_to);
+       if(relation_via!=NO_NODE_ID) logerror_node(relation_via);
+       if(relation_from!=NO_WAY_ID) logerror_way(relation_from);
+       logerror("Turn Relation %"Prelation_t" has more than one 'to' Way (used first one).\n",logerror_relation(relation_id));
+      }
+    if(relation_via_count>1)
+      {
+       if(relation_to!=NO_WAY_ID) logerror_way(relation_to);
+       if(relation_via!=NO_NODE_ID) logerror_node(relation_via);
+       if(relation_from!=NO_WAY_ID) logerror_way(relation_from);
+       logerror("Turn Relation %"Prelation_t" has more than one 'via' Node (used first one).\n",logerror_relation(relation_id));
       }
 
     if(relation_from!=NO_WAY_ID && relation_to!=NO_WAY_ID && relation_via!=NO_NODE_ID)
@@ -1031,11 +1169,17 @@ static double parse_speed(way_t id,const char *k,const char *v)
    {
     while(isspace(*ev)) ev++;
 
+    if(*ev==0)
+       return(value);
+
+    if(!strcmp(ev,"km/h") || !strcmp(ev,"kph") || !strcmp(ev,"kmph"))
+       return(value);
+
     if(!strcmp(ev,"mph"))
        return(1.609*value);
 
-    if(*ev==0 || !strcmp(ev,"kph"))
-       return(value);
+    if(!strcmp(ev,"knots"))
+       return(1.852*value);
 
     logerror("Way %"Pway_t" has an un-parseable tag '%s' = '%s' (after tagging rules); ignoring it.\n",logerror_way(id),k,v);
    }
@@ -1067,12 +1211,15 @@ static double parse_weight(way_t id,const char *k,const char *v)
    {
     while(isspace(*ev)) ev++;
 
+    if(*ev==0)
+       return(value);
+
     if(!strcmp(ev,"kg"))
        return(value/1000.0);
 
-    if(*ev==0 || !strcmp(ev,"T") || !strcmp(ev,"t")
-              || !strcmp(ev,"ton") || !strcmp(ev,"tons")
-              || !strcmp(ev,"tonne") || !strcmp(ev,"tonnes"))
+    if(!strcmp(ev,"T") || !strcmp(ev,"t") ||
+       !strcmp(ev,"ton") || !strcmp(ev,"tons") ||
+       !strcmp(ev,"tonne") || !strcmp(ev,"tonnes"))
        return(value);
 
     logerror("Way %"Pway_t" has an un-parseable tag '%s' = '%s' (after tagging rules); ignoring it.\n",logerror_way(id),k,v);
@@ -1106,6 +1253,26 @@ static double parse_length(way_t id,const char *k,const char *v)
     int en=0;
     int feet=0,inches=0;
 
+    while(isspace(*ev)) ev++;
+
+    if(*ev==0)
+       return(value);
+
+    if(!strcmp(ev,"m") || !strcmp(ev,"metre") || !strcmp(ev,"metres") || !strcmp(ev,"meter") || !strcmp(ev,"meters"))
+       return(value);
+
+    if(!strcmp(ev,"'"))
+       return(value*0.254);
+
+    if(!strcmp(ev,"′"))
+       return(value*0.254);
+
+    if(!strcmp(ev,"’"))
+       return(value*0.254);
+
+    if(!strcmp(ev,"ft") || !strcmp(ev,"feet"))
+       return(value*0.254);
+
     if(sscanf(v,"%d' %d\"%n",&feet,&inches,&en)==2 && en && !v[en])
        return((feet+(double)inches/12.0)*0.254);
 
@@ -1115,6 +1282,33 @@ static double parse_length(way_t id,const char *k,const char *v)
     if(sscanf(v,"%d'-%d\"%n",&feet,&inches,&en)==2 && en && !v[en])
        return((feet+(double)inches/12.0)*0.254);
 
+    if(sscanf(v,"%d' %d''%n",&feet,&inches,&en)==2 && en && !v[en])
+       return((feet+(double)inches/12.0)*0.254);
+
+    if(sscanf(v,"%d'%d''%n",&feet,&inches,&en)==2 && en && !v[en])
+       return((feet+(double)inches/12.0)*0.254);
+
+    if(sscanf(v,"%d'-%d''%n",&feet,&inches,&en)==2 && en && !v[en])
+       return((feet+(double)inches/12.0)*0.254);
+
+    if(sscanf(v,"%d′ %d″%n",&feet,&inches,&en)==2 && en && !v[en])
+       return((feet+(double)inches/12.0)*0.254);
+
+    if(sscanf(v,"%d′%d″%n",&feet,&inches,&en)==2 && en && !v[en])
+       return((feet+(double)inches/12.0)*0.254);
+
+    if(sscanf(v,"%d′-%d″%n",&feet,&inches,&en)==2 && en && !v[en])
+       return((feet+(double)inches/12.0)*0.254);
+
+    if(sscanf(v,"%d’ %d”%n",&feet,&inches,&en)==2 && en && !v[en])
+       return((feet+(double)inches/12.0)*0.254);
+
+    if(sscanf(v,"%d’%d”%n",&feet,&inches,&en)==2 && en && !v[en])
+       return((feet+(double)inches/12.0)*0.254);
+
+    if(sscanf(v,"%d’-%d”%n",&feet,&inches,&en)==2 && en && !v[en])
+       return((feet+(double)inches/12.0)*0.254);
+
     if(sscanf(v,"%d - %d%n",&feet,&inches,&en)==2 && en && !v[en])
        return((feet+(double)inches/12.0)*0.254);
 
@@ -1124,17 +1318,6 @@ static double parse_length(way_t id,const char *k,const char *v)
     if(sscanf(v,"%d feet %d inches%n",&feet,&inches,&en)==2 && en && !v[en])
        return((feet+(double)inches/12.0)*0.254);
 
-    if(!strcmp(ev,"'"))
-       return(feet*0.254);
-
-    while(isspace(*ev)) ev++;
-
-    if(!strcmp(ev,"ft") || !strcmp(ev,"feet"))
-       return(value*0.254);
-
-    if(*ev==0 || !strcmp(ev,"m") || !strcmp(ev,"metre") || !strcmp(ev,"metres"))
-       return(value);
-
     logerror("Way %"Pway_t" has an un-parseable tag '%s' = '%s' (after tagging rules); ignoring it.\n",logerror_way(id),k,v);
    }
 
