@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2013 Andrew M. Bishop
+ This file Copyright 2008-2017, 2023, 2025 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -22,10 +22,11 @@
 
 #include <stdio.h>
 #include <stdlib.h>
-#include <unistd.h>
 #include <string.h>
 #include <errno.h>
 
+#include "version.h"
+
 #include "types.h"
 #include "ways.h"
 
@@ -52,7 +53,7 @@
 char *option_tmpdirname=NULL;
 
 /*+ The amount of RAM to use for filesorting. +*/
-size_t option_filesort_ramsize=0;
+ssize_t option_filesort_ramsize=0;
 
 /*+ The number of threads to use for filesorting. +*/
 int option_filesort_threads=1;
@@ -88,7 +89,9 @@ int main(int argc,char** argv)
 
  for(arg=1;arg<argc;arg++)
    {
-    if(!strcmp(argv[arg],"--help"))
+    if(!strcmp(argv[arg],"--version"))
+       print_usage(-1,NULL,NULL);
+    else if(!strcmp(argv[arg],"--help"))
        print_usage(1,NULL,NULL);
     else if(!strncmp(argv[arg],"--dir=",6))
        dirname=&argv[arg][6];
@@ -108,6 +111,8 @@ int main(int argc,char** argv)
        option_loggable=1;
     else if(!strcmp(argv[arg],"--logtime"))
        option_logtime=1;
+    else if(!strcmp(argv[arg],"--logmemory"))
+       option_logmemory=1;
     else if(!strcmp(argv[arg],"--errorlog"))
        errorlog="error.log";
     else if(!strncmp(argv[arg],"--errorlog=",11))
@@ -155,19 +160,26 @@ int main(int argc,char** argv)
     print_usage(0,NULL,"Cannot use '--process-only' and filenames at the same time.");
 
  if(!option_filenames && !option_process_only)
-    print_usage(0,NULL,"File names must be specified unless using '--process-only'");
+    print_usage(0,NULL,"File names must be specified unless using '--process-only'.");
 
- if(!option_filesort_ramsize)
+ if(option_filesort_ramsize<0 || option_filesort_ramsize>1024*1024)
+    print_usage(0,NULL,"Sorting RAM size '--sort-ram-size=...' must be positive and in MB.");
+ else if(option_filesort_ramsize==0)
    {
 #if SLIM
-    option_filesort_ramsize=64*1024*1024;
-#else
     option_filesort_ramsize=256*1024*1024;
+#else
+    option_filesort_ramsize=1024*1024*1024;
 #endif
    }
  else
     option_filesort_ramsize*=1024*1024;
 
+#if defined(USE_PTHREADS) && USE_PTHREADS
+ if(option_filesort_threads<1 || option_filesort_threads>32)
+    print_usage(0,NULL,"Sorting threads '--sort-threads=...' must be small positive integer.");
+#endif
+
  if(!option_tmpdirname)
    {
     if(!dirname)
@@ -182,20 +194,26 @@ int main(int argc,char** argv)
       {
        if(!ExistsFile(tagging))
          {
-          fprintf(stderr,"Error: The '--tagging' option specifies a file that does not exist.\n");
+          fprintf(stderr,"Error: The '--tagging' option specifies a file '%s' that does not exist.\n",tagging);
           exit(EXIT_FAILURE);
          }
       }
     else
       {
-       if(ExistsFile(FileName(dirname,prefix,"tagging.xml")))
-          tagging=FileName(dirname,prefix,"tagging.xml");
-       else if(ExistsFile(FileName(DATADIR,NULL,"tagging.xml")))
-          tagging=FileName(DATADIR,NULL,"tagging.xml");
-       else
+       tagging=FileName(dirname,prefix,"tagging.xml");
+
+       if(!ExistsFile(tagging))
          {
-          fprintf(stderr,"Error: The '--tagging' option was not used and the default 'tagging.xml' does not exist.\n");
-          exit(EXIT_FAILURE);
+          char *defaulttagging=FileName(ROUTINO_DATADIR,NULL,"tagging.xml");
+
+          if(!ExistsFile(defaulttagging))
+            {
+             fprintf(stderr,"Error: The '--tagging' option was not used and the files '%s' and '%s' do not exist.\n",tagging,defaulttagging);
+             exit(EXIT_FAILURE);
+            }
+
+          free(tagging);
+          tagging=defaulttagging;
          }
       }
 
@@ -247,6 +265,12 @@ if(!option_process_only)
          *p=0;
         }
 
+      if((p=strstr(filename,".xz")) && !strcmp(p,".xz"))
+        {
+         fd=Uncompress_Xz(fd);
+         *p=0;
+        }
+
       if(option_changes)
         {
          printf("\nParse OSC Data [%s]\n==============\n\n",filename);
@@ -254,17 +278,18 @@ if(!option_process_only)
 
          if((p=strstr(filename,".pbf")) && !strcmp(p,".pbf"))
            {
-            logassert(0,"Unable to read a PBF file to apply changes (format does not permit this)");
+            fprintf(stderr,"Error: Unable to read a PBF file to apply changes (format does not permit this).\n");
+            exit(EXIT_FAILURE);
            }
          else if((p=strstr(filename,".o5c")) && !strcmp(p,".o5c"))
            {
             if(ParseO5CFile(fd,OSMNodes,OSMWays,OSMRelations))
-               exit(EXIT_FAILURE);
+               exit(EXIT_FAILURE); /* no message because one printed in parser function */
            }
          else
            {
             if(ParseOSCFile(fd,OSMNodes,OSMWays,OSMRelations))
-               exit(EXIT_FAILURE);
+               exit(EXIT_FAILURE); /* no message because one printed in parser function */
            }
         }
       else
@@ -275,17 +300,17 @@ if(!option_process_only)
          if((p=strstr(filename,".pbf")) && !strcmp(p,".pbf"))
            {
             if(ParsePBFFile(fd,OSMNodes,OSMWays,OSMRelations))
-               exit(EXIT_FAILURE);
+               exit(EXIT_FAILURE); /* no message because one printed in parser function */
            }
          else if((p=strstr(filename,".o5m")) && !strcmp(p,".o5m"))
            {
             if(ParseO5MFile(fd,OSMNodes,OSMWays,OSMRelations))
-               exit(EXIT_FAILURE);
+               exit(EXIT_FAILURE); /* no message because one printed in parser function */
            }
          else
            {
             if(ParseOSMFile(fd,OSMNodes,OSMWays,OSMRelations))
-               exit(EXIT_FAILURE);
+               exit(EXIT_FAILURE); /* no message because one printed in parser function */
            }
         }
 
@@ -307,7 +332,7 @@ if(!option_process_only)
     FreeWayList(OSMWays,1);
     FreeRelationList(OSMRelations,1);
 
-    return(0);
+    exit(EXIT_SUCCESS);
    }
 
 
@@ -329,21 +354,19 @@ if(!option_process_only)
  printf("\nProcess OSM Data\n================\n\n");
  fflush(stdout);
 
- /* Remove non-highway nodes by looking through the ways */
+ /* Remove non-highway nodes by looking through the ways (must be before splitting the ways) */
 
  RemoveNonHighwayNodes(OSMNodes,OSMWays,option_keep||option_changes);
 
- /* Separate the segments and way names and sort them. */
+ /* Separate the segments and way names and sort them (must be before indexing the segments) */
 
  OSMSegments=SplitWays(OSMWays,OSMNodes,option_keep||option_changes);
 
  SortWayNames(OSMWays);
 
- SortSegmentList(OSMSegments);
+ SortSegmentList(OSMSegments,OSMNodes,OSMWays);
 
- /* Process the segments and index them */
-
- ProcessSegments(OSMSegments,OSMNodes,OSMWays);
+ /* Index the segments (must be before processing relations) */
 
  IndexSegments(OSMSegments,OSMNodes,OSMWays);
 
@@ -353,42 +376,78 @@ if(!option_process_only)
 
  ProcessTurnRelations(OSMRelations,OSMNodes,OSMSegments,OSMWays,option_keep||option_changes);
 
- /* Compact the ways (must be after processing turn relations) */
+ /* Compact the ways */
 
  CompactWayList(OSMWays,OSMSegments);
 
- /* Index the segments */
+ /* Sort the nodes and segments geographically */
+
+ SortNodeListGeographically(OSMNodes);
+
+ SortSegmentListGeographically(OSMSegments,OSMNodes);
+
+ /* Re-index the segments */
 
  IndexSegments(OSMSegments,OSMNodes,OSMWays);
 
- /* Prune unwanted nodes/segments. */
+ /* Sort the turn relations geographically */
+
+ SortTurnRelationListGeographically(OSMRelations,OSMNodes,OSMSegments,0);
+
+ /* Prune unwanted nodes/segments */
 
  if(option_prune_straight || option_prune_isolated || option_prune_short)
    {
     printf("\nPrune Unneeded Data\n===================\n\n");
     fflush(stdout);
 
-    StartPruning(OSMNodes,OSMSegments,OSMWays);
-
     if(option_prune_straight)
+      {
+       StartPruning(OSMNodes,OSMSegments,OSMWays);
+
        PruneStraightHighwayNodes(OSMNodes,OSMSegments,OSMWays,option_prune_straight);
 
+       FinishPruning(OSMNodes,OSMSegments,OSMWays);
+
+       RemovePrunedNodes(OSMNodes,OSMSegments);
+       RemovePrunedSegments(OSMSegments,OSMWays);
+       CompactWayList(OSMWays,OSMSegments);
+       RemovePrunedTurnRelations(OSMRelations,OSMNodes);
+
+       IndexSegments(OSMSegments,OSMNodes,OSMWays);
+      }
+
     if(option_prune_isolated)
+      {
+       StartPruning(OSMNodes,OSMSegments,OSMWays);
+
        PruneIsolatedRegions(OSMNodes,OSMSegments,OSMWays,option_prune_isolated);
 
+       FinishPruning(OSMNodes,OSMSegments,OSMWays);
+
+       RemovePrunedNodes(OSMNodes,OSMSegments);
+       RemovePrunedSegments(OSMSegments,OSMWays);
+       CompactWayList(OSMWays,OSMSegments);
+       RemovePrunedTurnRelations(OSMRelations,OSMNodes);
+
+       IndexSegments(OSMSegments,OSMNodes,OSMWays);
+      }
+
     if(option_prune_short)
+      {
+       StartPruning(OSMNodes,OSMSegments,OSMWays);
+
        PruneShortSegments(OSMNodes,OSMSegments,OSMWays,option_prune_short);
 
-    FinishPruning(OSMNodes,OSMSegments,OSMWays);
+       FinishPruning(OSMNodes,OSMSegments,OSMWays);
 
-    /* Remove the pruned nodes, segments, ways and relations and update the indexes */
+       RemovePrunedNodes(OSMNodes,OSMSegments);
+       RemovePrunedSegments(OSMSegments,OSMWays);
+       CompactWayList(OSMWays,OSMSegments);
+       RemovePrunedTurnRelations(OSMRelations,OSMNodes);
 
-    RemovePrunedNodes(OSMNodes,OSMSegments);
-    RemovePrunedSegments(OSMSegments,OSMWays);
-    CompactWayList(OSMWays,OSMSegments);
-    RemovePrunedTurnRelations(OSMRelations,OSMNodes);
-
-    IndexSegments(OSMSegments,OSMNodes,OSMWays);
+       IndexSegments(OSMSegments,OSMNodes,OSMWays);
+      }
    }
 
  /* Repeated iteration on Super-Nodes and Super-Segments */
@@ -416,6 +475,10 @@ if(!option_process_only)
       {
        SegmentsX *SuperSegments2;
 
+       /* Index the super-segments */
+
+       IndexSegments(SuperSegments,OSMNodes,OSMWays);
+
        /* Select the super-nodes */
 
        ChooseSuperNodes(OSMNodes,SuperSegments,OSMWays);
@@ -435,10 +498,6 @@ if(!option_process_only)
 
     DeduplicateSuperSegments(SuperSegments,OSMWays);
 
-    /* Index the segments */
-
-    IndexSegments(SuperSegments,OSMNodes,OSMWays);
-
     /* Check for end condition */
 
     if(SuperSegments->number==nsuper)
@@ -483,7 +542,7 @@ if(!option_process_only)
 
  /* Sort the turn relations geographically */
 
- SortTurnRelationListGeographically(OSMRelations,OSMNodes,OSMSegments);
+ SortTurnRelationListGeographically(OSMRelations,OSMNodes,OSMSegments,1);
 
  /* Output the results */
 
@@ -506,6 +565,10 @@ if(!option_process_only)
 
  SaveRelationList(OSMRelations,FileName(dirname,prefix,"relations.mem"));
 
+ /* Free the memory (delete the temporary files) */
+
+ FreeSegmentList(OSMSegments);
+
  /* Close the error log file and process the data */
 
  if(errorlog)
@@ -537,18 +600,19 @@ if(!option_process_only)
  FreeWayList(OSMWays,0);
  FreeRelationList(OSMRelations,0);
 
- FreeSegmentList(OSMSegments);
+ printf("\n");
+ fflush(stdout);
 
  printf_program_end();
 
- return(0);
+ exit(EXIT_SUCCESS);
 }
 
 
 /*++++++++++++++++++++++++++++++++++++++
   Print out the usage information.
 
-  int detail The level of detail to use - 0 = low, 1 = high.
+  int detail The level of detail to use: -1 = just version number, 0 = low detail, 1 = full details.
 
   const char *argerr The argument that gave the error (if there is one).
 
@@ -557,48 +621,64 @@ if(!option_process_only)
 
 static void print_usage(int detail,const char *argerr,const char *err)
 {
- fprintf(stderr,
-         "Usage: planetsplitter [--help]\n"
-         "                      [--dir=<dirname>] [--prefix=<name>]\n"
+ if(detail<0)
+   {
+    fprintf(stderr,
+            "Routino version " ROUTINO_VERSION " " ROUTINO_URL ".\n"
+            );
+   }
+
+ if(detail>=0)
+   {
+    fprintf(stderr,
+            "Usage: planetsplitter [--version]\n"
+            "                      [--help]\n"
+            "                      [--dir=<dirname>] [--prefix=<name>]\n"
 #if defined(USE_PTHREADS) && USE_PTHREADS
-         "                      [--sort-ram-size=<size>] [--sort-threads=<number>]\n"
+            "                      [--sort-ram-size=<size>] [--sort-threads=<number>]\n"
 #else
-         "                      [--sort-ram-size=<size>]\n"
+            "                      [--sort-ram-size=<size>]\n"
 #endif
-         "                      [--tmpdir=<dirname>]\n"
-         "                      [--tagging=<filename>]\n"
-         "                      [--loggable] [--logtime]\n"
-         "                      [--errorlog[=<name>]]\n"
-         "                      [--parse-only | --process-only]\n"
-         "                      [--append] [--keep] [--changes]\n"
-         "                      [--max-iterations=<number>]\n"
-         "                      [--prune-none]\n"
-         "                      [--prune-isolated=<len>]\n"
-         "                      [--prune-short=<len>]\n"
-         "                      [--prune-straight=<len>]\n"
-         "                      [<filename.osm> ... | <filename.osc> ...\n"
-         "                       | <filename.pbf> ...\n"
-         "                       | <filename.o5m> ... | <filename.o5c> ..."
+            "                      [--tmpdir=<dirname>]\n"
+            "                      [--tagging=<filename>]\n"
+            "                      [--loggable] [--logtime] [--logmemory]\n"
+            "                      [--errorlog[=<name>]]\n"
+            "                      [--parse-only | --process-only]\n"
+            "                      [--append] [--keep] [--changes]\n"
+            "                      [--max-iterations=<number>]\n"
+            "                      [--prune-none]\n"
+            "                      [--prune-isolated=<len>]\n"
+            "                      [--prune-short=<len>]\n"
+            "                      [--prune-straight=<len>]\n"
+            "                      [<filename.osm> ... | <filename.osc> ...\n"
+            "                       | <filename.pbf> ...\n"
+            "                       | <filename.o5m> ... | <filename.o5c> ..."
 #if defined(USE_BZIP2) && USE_BZIP2
-         "\n                       | <filename.(osm|osc|o5m|o5c).bz2> ..."
+            "\n                       | <filename.(osm|osc|o5m|o5c).bz2> ..."
 #endif
 #if defined(USE_GZIP) && USE_GZIP
-         "\n                       | <filename.(osm|osc|o5m|o5c).gz> ..."
+            "\n                       | <filename.(osm|osc|o5m|o5c).gz> ..."
 #endif
-         "]\n");
+#if defined(USE_XZ) && USE_XZ
+            "\n                       | <filename.(osm|osc|o5m|o5c).xz> ..."
+#endif
+            "]\n");
 
- if(argerr)
+    if(argerr)
+       fprintf(stderr,
+               "\n"
+               "Error with command line parameter: %s\n",argerr);
+
+    if(err)
+       fprintf(stderr,
+               "\n"
+               "Error: %s\n",err);
+   }
+
+ if(detail==1)
     fprintf(stderr,
             "\n"
-            "Error with command line parameter: %s\n",argerr);
-
- if(err)
-    fprintf(stderr,
-            "\n"
-            "Error: %s\n",err);
-
- if(detail)
-    fprintf(stderr,
+            "--version                 Print the version of Routino.\n"
             "\n"
             "--help                    Prints this information.\n"
             "\n"
@@ -607,9 +687,9 @@ static void print_usage(int detail,const char *argerr,const char *err)
             "\n"
             "--sort-ram-size=<size>    The amount of RAM (in MB) to use for data sorting\n"
 #if SLIM
-            "                          (defaults to 64MB otherwise.)\n"
-#else
             "                          (defaults to 256MB otherwise.)\n"
+#else
+            "                          (defaults to 1024MB otherwise.)\n"
 #endif
 #if defined(USE_PTHREADS) && USE_PTHREADS
             "--sort-threads=<number>   The number of threads to use for data sorting.\n"
@@ -621,10 +701,11 @@ static void print_usage(int detail,const char *argerr,const char *err)
             "--tagging=<filename>      The name of the XML file containing the tagging rules\n"
             "                          (defaults to 'tagging.xml' with '--dir' and\n"
             "                           '--prefix' options or the file installed in\n"
-            "                           '" DATADIR "').\n"
+            "                           '" ROUTINO_DATADIR "').\n"
             "\n"
             "--loggable                Print progress messages suitable for logging to file.\n"
             "--logtime                 Print the elapsed time for each processing step.\n"
+            "--logmemory               Print the max allocated/mapped memory for each step.\n"
             "--errorlog[=<name>]       Log parsing errors to 'error.log' or the given name\n"
             "                          (the '--dir' and '--prefix' options are applied).\n"
             "\n"
@@ -655,6 +736,9 @@ static void print_usage(int detail,const char *argerr,const char *err)
 #endif
 #if defined(USE_GZIP) && USE_GZIP
             "                          Filenames ending '.gz' will be gzip uncompressed.\n"
+#endif
+#if defined(USE_XZ) && USE_XZ
+            "                          Filenames ending '.xz' will be xz uncompressed.\n"
 #endif
             "\n"
             "<transport> defaults to all but can be set to:\n"
