@@ -5,7 +5,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2014 Andrew M. Bishop
+ This file Copyright 2008-2016 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -68,7 +68,7 @@ struct _Segments
 
 #if !SLIM
 
- void        *data;             /*+ The memory mapped data. +*/
+ char        *data;             /*+ The memory mapped data. +*/
 
  Segment     *segments;         /*+ An array of segments. +*/
 
@@ -182,10 +182,12 @@ CACHE_DELETECACHE_PROTO(Segment)
 CACHE_FETCHCACHE_PROTO(Segment)
 CACHE_INVALIDATECACHE_PROTO(Segment)
 
+/* Data type */
+
+CACHE_STRUCTURE(Segment)
 
 /* Inline functions */
 
-CACHE_STRUCTURE(Segment)
 CACHE_NEWCACHE(Segment)
 CACHE_DELETECACHE(Segment)
 CACHE_FETCHCACHE(Segment)
@@ -226,7 +228,7 @@ static inline Segment *LookupSegment(Segments *segments,index_t index,int positi
 
 static inline index_t IndexSegment(Segments *segments,Segment *segmentp)
 {
- int position1=segmentp-&segments->cached[0];
+ int position1=(int)(segmentp-&segments->cached[0]);
 
  return(segments->incache[position1]);
 }
@@ -246,7 +248,7 @@ static inline index_t IndexSegment(Segments *segments,Segment *segmentp)
 
 static inline Segment *NextSegment(Segments *segments,Segment *segmentp,index_t node)
 {
- int position=segmentp-&segments->cached[-1];
+ int position=(int)(segmentp-segments->cached)+1;
 
  if(segmentp->node1==node)
    {
