@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2009-2015, 2017, 2019, 2023 Andrew M. Bishop
+ This file Copyright 2009-2015, 2017, 2019, 2023, 2025 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -921,7 +921,7 @@ static index_t filesort_merge(int fd_out,int nfiles,size_t itemsize,size_t large
 
     fds[i]=ReOpenFileBuffered(filename);
 
-    DeleteFile(filename);
+    DeleteFileBuffered(filename);
    }
 
  /* Fill the heap to start with */
