@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2014 Andrew M. Bishop
+ This file Copyright 2008-2014, 2019 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -58,11 +58,11 @@
 #define NO_LATLONG     ((latlong_t)0x80000000)
 
 
-/*+ The lowest number allowed for a fake node. +*/
-#define NODE_FAKE      ((index_t)0xffff0000)
+/*+ The lowest number allowed for a fake node (allows 65535 of them). +*/
+#define NODE_FAKE      (((index_t)~0)<<16)
 
-/*+ The lowest number allowed for a fake segment. +*/
-#define SEGMENT_FAKE   ((index_t)0xffff0000)
+/*+ The lowest number allowed for a fake segment (allows 65535 of them). +*/
+#define SEGMENT_FAKE   (((index_t)~0)<<16)
 
 
 /*+ The latitude and longitude conversion factor from floating point (radians) to integer. +*/
@@ -239,13 +239,14 @@ typedef enum _Highway
 
   Highway_Count        = 14,       /* One more than the number of highway types. */
 
-  Highway_OneWay       =  32,
-  Highway_Roundabout   =  64,
-  Highway_Area         = 128
+  Highway_CycleBothWays =  16,
+  Highway_OneWay        =  32,
+  Highway_Roundabout    =  64,
+  Highway_Area          = 128
  }
  Highway;
 
-#define HIGHWAY(xx) ((xx)&0x1f)
+#define HIGHWAY(xx) ((xx)&0x0f)
 
 /*+ A bitmask of multiple highway types. +*/
 typedef uint16_t highways_t;
@@ -338,9 +339,8 @@ typedef enum _Property
   Property_Tunnel       = 4,
   Property_FootRoute    = 5,
   Property_BicycleRoute = 6,
-  Property_CycleBothWays= 7,
 
-  Property_Count        = 8       /* One more than the number of property types. */
+  Property_Count        = 7       /* One more than the number of property types. */
  }
  Property;
 
@@ -361,7 +361,6 @@ typedef enum _Properties
   Properties_Tunnel       = PROPERTIES(Property_Tunnel       ),
   Properties_FootRoute    = PROPERTIES(Property_FootRoute    ),
   Properties_BicycleRoute = PROPERTIES(Property_BicycleRoute ),
-  Properties_CycleBothWays= PROPERTIES(Property_CycleBothWays),
 
   Properties_ALL          = PROPERTIES(Property_Count        )-1
  }
@@ -450,7 +449,7 @@ const char *TransportName(Transport transport);
 const char *PropertyName(Property property);
 
 const char *HighwaysNameList(highways_t highways);
-const char *AllowedNameList(transports_t allowed);
+const char *TransportsNameList(transports_t transports);
 const char *PropertiesNameList(properties_t properties);
 
 const char *HighwayList(void);
